#! /usr/bin/perl -w
package CompCheck;

use IPC::Open3;
use setup;

sub messname {
#returns a printing name for the code from a .ccf[tc] full name
# form:  code w/o ext (ccf file w/o .ccf)
my($ccf);
$ccf = shift;
my($ext,$name,@break);
@break = split(/\./,$ccf);  
$ext = $break[1]; #assume at most one "." (check in check1 below)
#print "$ccf: $ext\n"; #debug
if ($ccf =~ /^theory/ || $ext eq "ccft" || $ext eq "ccfc") { #not a .ccf file at all
  return $break[0];
}
open(CCF,"< $ccf") || die "no file $ccf to find a print name for";
$nameline = <CCF>;
@nameand = split(' ',$nameline);
$name = $nameand[0];
close(CCF);
if ($break[0] eq $name) {
  return $name;
}
return $name."(".$break[0].")";
}

sub killdead {
my($pid) = shift;
    #for MSW, no CHLD signal, so $dead is always 0
    # kill(0,) is accurate, but only if some loss of control
    # (but not sleep!) occurs first, and kill() itself is such a loss,
    #  so do it twice
    # But kill(0,) calls zombies alive, so fails on UNIX...
if (kill(0, $pid)) {
  $d = 0;
} else {
  $d = 1;
}  #phony call for MSW
if (kill(0, $pid)) {
  $dd = 0;
} else {
  $dd = 1;
}  
#print "!kill-1: $d; !kill-2: $dd; dead: $dead\n";  #debug
return $dd;  #second call value
}

# (Flag, Out) = ckblock(FD, Code, MidMsg)
# returns Out from reading FD in Code, with Flag 0.  
#  If can't read or not numeric, etc, Flag 1 and Out may be undefined. 
#   Prints warning messages for problems, with string MidMsg.
sub ckblock {
my($fd) = shift;
my($prog) = shift;
my($msg) = shift;
my($Out);
$bits = "";
vec($bits,fileno($fd),1) = 1;
if (select($bits,undef,undef,2)) {#output ready by 2 sec (sometimes lies)
  $Out = <$fd>;  #read its output
  unless (defined($Out)) { #MSW or STDERR does this
    warn "  code $prog has no $msg output\n";
    return 1, undef;
  }
  chomp($Out);
  if ($Out !~ /$numPat/o) {
    warn "  code $prog has non-numeric $msg output: $Out\n";
    return 1, $Out;
  }
} else { #too long a wait for output
  warn "  code $prog is slow with $msg output\n";
  warn "   (maybe buffered and $prog hung)\n";
  return 1, undef;
}
return 0, $Out;  #successful
}

sub check1 {
#check a ccf file for proper format & proper names
# also run the code to detect proper sequence of operations
#  if execution will be needed in COMP
my($comp,$V); 
my($prog, $Wrong, $line, @bits, $stateless, $FirstFirst, $SecondFirst,
  $AnyRight, $num_isubd, $num_ssubd, $LeftI, $LeftS, $RightI, $RightS); 
$comp = shift; # .ccf file name
$V = shift; #true for verbose
open(CCF, $comp) || die "could not open the file ", $comp;
$progline = <CCF> ; # get the code line
@codeand = split(' ',$progline);
$prog = $codeand[0];
$stateless = 0; #does not include concurrent
if (not defined($codeand[1])) {
  $stateless = 1;
  $codeand[1] = "" 
}
$codetype = $codeand[1];
$concurrent = 0;
$stateless = 0 if ($codetype eq "state");
$concurrent = 1 if ($codetype eq "concurrent");
#read & check .ccf file
$Wrong = 0;  #if TRUE, found something wrong with .ccf file
# Wrong==2 means a problem that precludes some further checking
#ccf names must have at most one "." and no other specials, no blanks
#code name no extension, no specials or blanks
if ($comp =~ /\ /) {
  warn "the name '$comp' contains a blank character\n";
  $Wrong = 1;
}
@bits = split(/\./,$comp);
if (defined($bits[2])) {
  warn "$comp has too many '.'\n";
  $Wrong = 1;
}
if ($bits[1] ne "ccf") {
  warn "$comp must have extension '.ccf'\n";
  $Wrong = 1;
}
if ($bits[0] =~ /\W/) {
  warn "the name '$bits[0]' contains some special character\n";
  $Wrong = 1;
}

if ($prog =~ /\ /) {
  warn "the name '$prog' contains a blank character\n";
  $Wrong = 1;
}
@bits = split(/\./,$prog);
if (defined($bits[1])) {
  warn "code file name $prog should not have an extension\n";
  #$Wrong = 1;
}
if ($bits[0] =~ /\W/) {
  warn "the name '$bits[0]' contains some special character\n";
  $Wrong = 1;
}
$FirstFirst = 1;
$SecondFirst = 0;
$AnyRight = 0;  #no value is really OK if first line is blank...
$num_isubd = 0;
$num_ssubd = 0; #also used to count 2nd input on concurrent
$beforeblank = 1;
ALLL:
while ($line = <CCF>) {
  chomp($line);
  @bits = split(" ",$line);
  #print $bits[2];  #debug
  if ($beforeblank) {
    $num_isubd++;
  } else {
    $num_ssubd++;
  }
  if ($FirstFirst && $line ne "") { #left-most input
    $FirstFirst = 0;
    $LeftI = $bits[0];
    $AnyRight = $LeftI;
  }
  if ($SecondFirst && $line ne "") { # left-most state
    $SecondFirst = 0;
    $LeftS = $bits[0];
    $AnyRight = $LeftS;
  }
  if ($line eq "") {
    if ($stateless) {
      warn "  stateless components may not have two subdomain sets\n";
      $Wrong = 2;
      last ALLL;  #abandon further checks
    }
    unless ($beforeblank) { #another blank!
      warn "  contains extra blank line(s) near [ , $AnyRight)\n";
      $Wrong = 2;
      last ALLL;  #abandon further checks
    }
    $beforeblank = 0;
    $SecondFirst = 1; 
    $num_isubd--; #subtract the blank line
  } elsif ($bits[0] > $bits[1]) {
    warn "  Line ",qq{$line}," has empty interval [",$bits[0],",",$bits[1],")\n";
    $Wrong = 2;  #but keep going...
  } elsif ($prog !~ /^theory/ && ((not defined($bits[2])) 
           || ($beforeblank && $bits[2] == 0) )) {
    warn "  Sampling count missing on inteval [",$bits[0],",",$bits[1],")\n";
    $Wrong = 1;
  }
  if ($Wrong < 2 && $line ne "") { #OK to make check for interval coverage
    unless ($AnyRight == $bits[0]) {
      warn "  Intervals overlap or are not contiguous near [$bits[0],$$bits[1])\n";
#print "$line\n";  #debug
      $Wrong = 2;
    } 
  }
  $AnyRight = $bits[1];
  if ($beforeblank) {
    $RightI = $AnyRight;
  } else {
    $RightS = $AnyRight;
  }
} #end while going through file
close (CCF);
if (!$stateless && $beforeblank) { #no break where should be
  warn "  Components with state or concurrency must have two subdomain sets\n";
  $Wrong = 2;
}
unless ($prog =~ /^theory/ || -e $prog) {
  warn "  No code(?) file ",$prog," for ",$comp,"\n";
  $Wrong = 1;
} 
if (-e $prog && !$Wrong) { #hard checks on the code, only if names OK
  if ($stateless) {
    $compt = $comp."t";
  } else {
    $compt = $comp."c";
  }
  if (!-e $compt || (-M $compt >= -M $prog)) { #code is newer, will have to execute 
    @base = split(/\./, $prog);
    $basename = $base[0];
    $stateName = "$basename.state";
    unlink $stateName;
    #execute: try input at middle of domain, check for numerical output and run time 
    # and state if any, and blocking sequence
    $midpt = ($LeftI + $RightI)/2.0;
    $numPat = '^\s*([-+]?)(?=\d|\.\d)\d*(\.\d*)?([Ee]([+-]?\d+))?(\r)?\s*$';  #C-style floats, maybe CR at end
    @pre = split(' ',$prog);
    unless (setup::isitMSW() || -X $prog || $pre[0] eq "perl") {
      warn "  code file $prog is not executable\n";
      $Wrong = 1;
      goto JUMP;
    }
    local $SIG{CHLD} = sub{$dead = 1;};
    $dead = 0; 
    $pid = open3(\*CIN,\*COUT,\*CERR,$prog);  #never fails?!
#print "open3 pid:  $pid for $prog\n";  #debug
    sleep(1); #to allow child to die if it's not blocking
    if (killdead($pid) || $dead) {# chile is dead by some measure
      warn "  code $prog does not read input\n";
      goto EJUMP;
    }
    print CIN "$midpt\n";  # send it input
    ($err, $out) = ckblock(COUT, $prog, "");
    goto EJUMP if ($err);
    ($err, $out) = ckblock(CERR, $prog, "non-functional");
    goto EJUMP if ($err);
    if ($concurrent) { #check further pattern:  read/out/err/err
      sleep(1);
      if (killdead($pid) || $dead) {
        warn "  code $prog does not read parallel input\n";
        goto EJUMP;
      }
      $midpt = ($LeftS + $RightS)/2.0;
      print CIN "$midpt\n";  # send parallel input;
      ($err, $out) = ckblock(COUT, $prog, "parallel");
      goto EJUMP if ($err);
      ($err, $out) = ckblock(CERR, $prog, "parallel non-functional");
      goto EJUMP if ($err);
      ($err, $out) = ckblock(CERR, $prog, "final parallel non-functional");
      goto EJUMP if ($err);
    }
    sleep(1); #child should die in there
    if (!killdead($pid) && !$dead) { #chile still running by all measures
      warn "  code $prog didn't terminate properly\n";
      goto EJUMP;
    }
    #passed all execution blocking-sequence checks!
    waitpid($pid,0);
    if (!$stateless && !$concurrent) { #check state file, too
      if (-e $stateName) {
        open(SN,$stateName);
        chomp($stateVal = <SN>);
        if ($stateVal !~ /$numPat/o) {
          warn "  State file contents non-numeric for $prog: $stateVal";
          $Wrong = 1;
        }
        close(SN);
      } else {
        warn "  No initial state file $prog.state created by $prog\n";
        $Wrong = 1;
      }
    }
    #should execute again to see if state is read and written when initialized...TBD
    unlink $stateName;
    #system("rm -f $stateName");
  }
} #end difficult code checks
goto JUMP;  # EJUMP code not to fall in
EJUMP: #error exit from code checks; code may be running with $pid
$Wrong = 1; #fatal mistake
sleep(1);
if (!killdead($pid) || !$dead) { #someone thinks it's still running
  kill(9,$pid);
}
waitpid($pid,0); #kill zombie if any
JUMP:
if ($V) {
  warn "Configuration file $comp with code file $prog:\n";
  if ($concurrent) {
    warn "  $num_ssubd parallel return subdomains on [$LeftS, $RightS) and $num_isubd input subdomains on [$LeftI, $RightI)\n";
  } elsif ($stateless) {
    warn "  stateless with $num_isubd input subdomains on [$LeftI, $RightI) \n";
  } else {
    warn "  has $num_ssubd state subdomains on [$LeftS, $RightS) and $num_isubd input subdomains on [$LeftI, $RightI)\n";
  }
}
if ($Wrong) {
  warn "Configuration/code file errors; fix $comp/$prog and try again\n";
  warn "--------------------------------\n";
} else {
  if ($V) {
    warn "Configuration file $comp seems OK\n";
  }
}
return ($stateless,$concurrent, !$Wrong);
} #end CompCheck

sub CheckIncr {
# comments TBD
my($comp);
my($newfile,$oldfile,$incremental,$dc,$dcl,$dcl1,$dcl2,@statvec,$oldTime,$newTime,
  $stillinput,$tl); 
$comp = shift;
$newfile = $comp;
$oldfile = $comp."b";
#must do it if oldfile exists yet newfile is more recent 
$incremental = 0; #most cases don't
if (-e $oldfile) {
  @statvec = stat($newfile);
  $newTime = $statvec[9];
  @statvec = stat($oldfile);
  $oldTime = $statvec[9];
  if ($newTime > $oldTime) { #new is more recent
    open(NEW,"<$newfile");
    open(OLD,"<$oldfile");  #these exist from above
    open(INCR,">$comp"."i") or die "can't create incremental file";
    $Nl = $Ol = 1; #first time thru
    IL:
    while ($Nl || $Ol) { #usually break out 
      $Nl = <NEW>;
      $Ol = <OLD>;
      if ($Nl && $Ol && $Nl eq $Ol) { #neither EOF, in sync and same line
        next IL;
      }
      #lines differ
      if ($Nl) { #real new line, not EOF 
        print INCR $Nl;
        $incremental = 1;
        #get back in sync
        $Nl = <NEW>;
        unless ($Nl) {
          last IL;
        }
        unless ($Ol) {
          while ($Nl = <NEW>) { #copy rest of new
            print INCR $Nl;
            }
          last IL;
        }
        @Ns = split(' ',$Nl);
        $NL = $Ns[0]; $NR = $Ns[1];
        @Os = split(' ',$Ol);
        $OL = $Os[0]; $OR = $Os[1];
        while ($OL != $NL) { #one is behind
          ILL:
          while ($OL < $NL) { #old behind
            $Ol = <OLD>;
            if ($Ol) {
              if ($Ol eq $NL) {#back in sync
                next IL;
              }
              @Os = split(' ',$Ol);
              $OL = $Os[0]; $OR = $Os[1];
              next ILL; 
            } else {
              print INCR $Nl;
              $incremental = 1;
              while ($Nl = <NEW>) { #copy rest of new 
                print INCR $Nl;
              }
              last IL;
            }
          }  #old not behind now
          ILLL:
          while ($NL < $OL) { #new behind
            print INCR $Nl;
            $incremental = 1;
            $Nl = <NEW>;
            if ($Nl) {
              if ($Ol eq $Nl) { #back in sync
                next IL
              }
              @Ns = split(' ',$Nl);
              $NL = $Ns[0]; $NR = $Ns[1];
              next ILLL;
            } else {
              last IL;
            }
          }
        } # end one is behind
        # must be same first element
        if ($Ol ne $Nl) { #in sync but differ
          print INCR $Nl;
          $incremental = 1;
        }
        next IL;
      } else { #end of new file, all done
        if ($Ol) { #old file is longer
          $incremental = 1;
        }
        last IL;
      }
    }
    close(INCR);
    close(NEW);
    close(OLD);

  }
}
return $incremental;
}

return 1;
