package component;

use IPC::Open3;
use setup;

#All production component executions are done here.  Checking for improper
# sequence, values, etc., is done in "CompCheck::check1" so here it is 
# assumed that the code to executed behaves itself.
#
#Executes and supplies input to a standard component w/o the object frills.
# (out, err) = excomp(in, comp, control, ...);
#  To avoid shell, make "comp" a single word
sub excomp{
my $in = shift;
my $code = shift;
my $control = shift;  #may be undefined (-> "none"), or "begin" "cont" "end"
my ($out,$run,$more);
  @code2 = split(' ',$code);
  unless (defined($control)) {
    $control = "none";
  }
  if ($control ne "end") { # param indicates other runs follow

    my $pid = shift;
    my $cin = shift;
    my $cout = shift;
    my $cerr = shift;  #descriptors passed back in
  }
  #unless ($code2[0] eq "perl" || setup::isitMSW() || -X $code) {
    #$out = "notexec";
    #$run = "notexec";
  #} else {
    if ($control eq "none" || $control eq "begin") {
#print "About to Open3...\n";  #debug
      #local $SIG{CHLD} = sub{$dead = 1;};
      #$dead = 0; #ready for signal from chile
      $pid = open3(\*CIN, \*COUT, \*CERR, $code);
#print "starting pid $pid for $code\n";  #debug
      #if ($dead) { #something killed child, maybe Perl syntax error
        #waitpid($pid,0); #kill zombie
        #return "notexec","notexec";
      #}
#print "Back from Open3...\n"; #debug
      $cin = CIN;
      $cout = COUT;
      $cerr = CERR;
    }
#print "$control $pid $cin (from component)\n";  #debug
    print $cin ("$in\n");
    if ($control eq "none" || $control eq "end") {
      close($cin);
    }
#print "sent it $in\n"; #debug
    if ($control ne "end") { #no output on termination
      chomp($out = <$cout>);
    #if (not defined($out)) { #probably not a program
      #$out = "undefined";
    #}
    #chomp($out);
      chomp($run = <$cerr>);
    #if (not defined($run)) { #probably not a program
      #$run = "undefined";
    #}
    #chomp($run);
#print "got back $out & $run\n"; #debug
    }
    if ($control eq "none" || $control eq "end") {
      close($cout);
      close($cerr);
      waitpid($pid,0);
    }
  #}
  if ($control eq "none") {
    return $out,$run;
  } else {
    return $out,$run,$pid,$cin,$cout,$cerr;
  }
}

sub start{
# (out, run) = start(input,code,index)
#  runs code down to first print, leaves descriptors open
#  index modifies names used
my $in = shift;
my $code = shift;
my $index = shift;
my ($out,$run);
$fin = "CIN$index";
$fout = "COUT$index";
$ferr = "CERR$index";
#local $SIG{CHLD} = sub{$dead = 1;};
#$dead = 0; #ready for signal from chile
${"pid$index"} = open3(*$fin, *$fout, *$ferr, $code);
#if ($dead) { #something killed child, maybe Perl syntax error
  #waitpid(${"pid$index"},0); #kill zombie
  #return "notexec","notexec";
#}
#print "started $code index $index with pid " .${"pid$index"}."\n"; #debug
print $fin ("$in\n");
chomp($out = <$fout>);
#print "$code out: $out\n"; #debug
chomp($run = <$ferr>);
#print "$code run: $run\n"; #debug
return ($out, $run);
}

sub finish{
# (out, run1, run2) = finish(input,index)
#  completes code with index 
#  run1 is the parallel time, run2 after
my $in = shift;
my $index = shift;
my ($out,$run1,$run2);
$fin = "CIN$index";
$fout = "COUT$index";
$ferr = "CERR$index";
#print "finishing index $index, pid ".${pid$index"}."\n";
print $fin ("$in\n");
close $fin;
chomp($out = <$fout>);
close $fout;
#print "$index out: $out\n";
chomp($run1 = <$ferr>);
chomp($run2 = <$ferr>);
#print "$index runs: $run1, $run2\n";
close $ferr;

waitpid(${"pid$index"},0);
return ($out, $run1, $run2);
}

# Constructs and initializes a new component object.
# Arguments:
#   component => "some_executible",
#   statefile => "some_file"    <-- optional
sub new {
  my $invocant = shift;
  my $class = ref($invocant) || $invocant;
  my $self = { @_ };
  die "No component specified" unless (defined($self->{component}));
  $Native = setup::fifo();
#$Native = 0;   #debug don't use fifo pipe (for MSW, always)
  unless (defined($self->{statefile})) {
    $self->{statefile} = $self->{component}.".state";
  }
  unless (defined($self->{comparg})) {
    ($self->{comparg}) = "";
  }
  $self->{first} = 1; #flag to start FIFO
  unless (defined($self->{stateless})) {
    $self->{stateless} = 0; #assume it uses state
  }
  unless (defined($self->{control})) {
    $self->{control} = "none";
  }
  return bless $self, $class;
}

# Runs the component passed to the constructor.
# Arguments:
#  $_[0] = input
#  $_[1] = state  (not always sent)
# Returns: 1 for success, 0 for failure
# On POSIX systems can use a FIFO pipe for state; otherwise use files
sub run {
  my $self = shift;
  my $sfile = $self->{statefile};
  my $input = shift;
   unless (defined($input)) { $input = $self->{input}; }
  my $state = shift;
  unless (defined($state)) {
    $state = $self->{state};
  }
  my $control = $self->{control};
  my ($pid, $wtr, $rdr, $err, $output, $runtime, @outstrm);

  if (-e $self->{statefile}) { # component has initialized with real file
    #make named pipe (but not Windows)
    if ($Native && $self->{first}) {
      unless ($self->{stateless}) {
        open(SW, "< $sfile") || return 0;
        chomp($state = <SW>);
        close SW;
        unlink $sfile;
#open (DB, ">>/tmp/comp");print DB "state read from initial real file:  $state\n";  #debug
        use POSIX 'mkfifo';
        mkfifo( $sfile, 0666 ) || return 0;
#open (DB, ">>/tmp/comp");print DB "made pipe $sfile \n";  #debug
        $self->{first} = 0;
      }
    }
  } else {
    $self->{first} = 1;  #(re)initialized by removing state file
  }

  if ($control eq "none" || $control eq "begin") {
    #local $SIG{CHLD} = sub{$dead = 1;};
    #$dead = 0; #ready for signal from chile
    unless ($self->{comparg}) {
      $pid = open3(\*WTR, \*RDR, \*ERR, $self->{component});
    } else {
      $pid = open3(\*WTR, \*RDR, \*ERR, $self->{component}, $self->{comparg});
    }
    #if ($dead) { #something killed child, maybe Perl syntax error
      #waitpid($pid,0); #kill zombie
      #return "notexec","notexec";
    #}
#die "failed to start $self{component}" if ($pid == 0);  #debug
    $self->{pid} = $pid;
    $self->{wtr} = WTR;
    $self->{rdr} = RDR;
    $self->{err} = ERR;
  } 
  # get into locals in any case
  $pid = $self->{pid};
  $wtr = $self->{wtr};
  $rdr = $self->{rdr};
  $err = $self->{err};
#open(DB, ">>/tmp/comp"); print DB "$self->{component}: pid=$pid, ctl=$control, in=$input\n";  #debug
  if ($pid) {
    print $wtr "$input\n";
#print "befor state stuff...\n";  #debug
    if ($Native && $control ne "end") { #use the pipe
      unless ($self->{stateless} || $self->{first}) { #don't do any piping
        # send state to the code, unblocking initial read
        open(STATEWRITER, ">", $sfile) || return 0;
        print STATEWRITER $state;
        close(STATEWRITER);
        # get state from code final write
        open(STATEREADER, "<", $sfile) || return 0;
        $state = <STATEREADER>;
        close(STATEREADER);
      }
    }
#print "aftr state stuff...\n";
    if ($control eq "end") { #shut down w/o waiting for i/o
      close($rdr);
      close($err);
      close($wtr);
      waitpid($pid, 0);  #wait for child to finish
    } else {
      $runtime = <$err>;
      #if ($runtime =~ /[A-Z]/) {
        #print STDERR "Component error: $runtime\n";
      #}
      $output = <$rdr>;
      #if ($output =~ /[A-Z]/) {
        #print STDERR "Component said: $output\n";
      #}
      if ($control eq "none") {
        close($rdr);
        close($err);
        close($wtr);
        waitpid($pid,0);
      }

      if ($Native) { #read the pipe the first time after initialization w/ real file
        if ($self->{first} && !($self->{stateless})) { 
          if (-e $sfile) { #component wrote a state file
            open(STATEREADER, "<$sfile") || return 0; #read real file first time
            $state = <STATEREADER>;
            close(STATEREADER);
          } else {
            $self->{stateless} = 1; #no state for sure
          }	 
        }
      } else { #always read file if state
        unless ($self->{stateless}) {
          open(STATEREADER, "< $self->{statefile}") || #return 0; 
           die "$self->{statefile} is gone!";
          $state = <STATEREADER>;
          close(STATEREADER);
        }
      }
      chomp($self->{output} = $output);
      chomp($self->{runtime} = $runtime);
      chomp($self->{state} = $state);
    } #end not "end" skip
     return 1;
  } else { #can't do the run
     return 0;
  }
}

sub NoState {
  my $self = shift;
  unlink $self->{statefile};
  return 1;
}

sub getState {
 my $self = shift;
 if ($self->{stateless}) {
   return 0;
 } else {
   return $self->{state};
 }
}

sub getOutput {
 my $self = shift;
 return $self->{output};
}

sub getRuntime {
 my $self = shift;
 return $self->{runtime};
}

return 1;
