package sampling; 

# only uses $globdata when module not being used in OO fashion
$globdata = 0;

#Initialize with init() for equispaced samples, with randomize() for
#  random sequences.
#  Then next() delivers the next point, reutrning "." at end.
#

# init - initialize equispaced sampling interval
# params:
# - interval begin - start of the interval
# - interval end - end of the interval
# - count - count number
sub init {

  # make this an OO compatible constructor

  my $param1 = shift;
  my $class = ref($param1);
#print "$param1 to sampling\n"; #debug
  if ((! $class) && !($param1 =~ /\d/)) {$class = $param1;}

  # get params

  if ($class) {
    $interval_begin = shift ;
  } else {
    $interval_begin = $param1;
  }
  $interval_end = shift ; 
  $sample_count = shift ; 

  # verify params

  die "Init() - interval_begin not defined.\n" if not defined $interval_begin ; 
  die "Init() - interval_end not defined.\n" if not defined $interval_end ; 
  die "Init() - count not defined.\n" if not defined $sample_count ;

  die "Init() - count must be non-negative value.\n" if ($sample_count < 0) ; 
  die "Init() - invalid interval boundary values.\n" if ($interval_begin >= $interval_end) ; 
  if ($sample_count == 1) { #midpoint
    #not used, see "next()" return
    #$ret_next_value = ($interval_begin + $interval_end)/2;
    #$interval_spacing = 2*$ret_next_value;
    #$interval_end = $ret_next_value;
  }
  elsif ($sample_count == 2) { #arcane!
    $interval_spacing = ($interval_end - $interval_begin)/4;
    $ret_next_value = $interval_begin + $interval_spacing;
    $interval_spacing *= 2;
  }
  else { # start at interval left
    $interval_spacing = ($interval_end - $interval_begin - .0000000001)/($sample_count-1);
    $ret_next_value = $interval_begin;
  }

  $data = {'count' => $sample_count,
           'sample_count' => $sample_count,
           'ret_next_value' => $ret_next_value,
           'interval_spacing' => $interval_spacing,
           'interval_end' => $interval_end,
           'interval_begin' => $interval_begin,
           'random' => 0 };

  return bless $data, $class if $class;
  $globdata = $data;
  return;
}

# randomize - initialize sampling interval for random sequences

# params:
# - interval begin - start of the interval
# - interval end - end of the interval
# - count - count number
sub randomize {
  # make this an OO compatible constructor

  my $param1 = shift;
  my $class = ref($param1);
#print "$param1 to sampling\n"; #debug
  if ((! $class) && !($param1 =~ /\d/)) {$class = $param1;}

  # get params

  if ($class) {
    $interval_begin = shift ;
  } else {
    $interval_begin = $param1;
  }
  $interval_end = shift ;
  $sample_count = shift ;

  # verify params

  die "randomize() - interval_begin not defined.\n" if not defined $interval_begin ;
  die "randomize() - interval_end not defined.\n" if not defined $interval_end ;
  die "randomize() - count not defined.\n" if not defined $sample_count ;

  die "randomize() - count must be positive value.\n" if ($sample_count <= 0) ;
  die "randomize() - invalid interval boundary values.\n" if ($interval_begin >= $interval_end) ;


  $ret_next_value = rand($interval_end - $interval_begin - .0000000001) + $interval_begin;

  $data = {'count' => $sample_count,
           'ret_next_value' => $ret_next_value,
           'interval_begin' => $interval_begin,
           'interval_end' => $interval_end,
           'random' => 1 };

  return bless $data, $class if $class;
  $globdata = $data;
  return;
}

sub next {
  my $save_this_ret;
  my $data = shift || $globdata;
  die "Not initialized; call init or randomize first" unless $data;
 
  if ($data->{'count'} <=0) { #no more
    return ".";
  }
  $save_this_ret = $data->{'ret_next_value'};
  $data->{'count'} --;
  if ($data->{'random'}) {
    $data->{'ret_next_value'} = rand($data->{'interval_end'} - $data->{'interval_begin'} - .0000000001) + $data->{'interval_begin'};
  } else {
    if ($data->{'sample_count'} == 1) { #midpoint
      #$save_this_ret = ($data->{'interval_begin'} + $data->{'interval_end'})/2.0
      return ($data->{'interval_begin'} + $data->{'interval_end'})/2.0
    }
    if ($data->{'sample_count'} >= 3 && $data->{'count'} == 0) { #last point
      if ($data->{'interval_spacing'} <= .0000000001) { #skip last
        return ".";
      }
      $save_this_ret = $data->{'interval_end'} - .0000000001;
    }
    $data->{'ret_next_value'} += $data->{'interval_spacing'};
  }
#print "$save_this_ret from [$data->{'interval_begin'},$data->{'interval_end'})\n";  #debug
  return $save_this_ret;
}
return 1;
