#! /usr/bin/perl -w
# 
#  Called to determine operating-system dependent parameters.
#  Perl special variable $^O contains a string with a name
#    for the platform on which Perl was compiled, for example
#    "MSWin32" for many Windows systems.  This name is entered
#    in a number of arrays below.  To control the behavior
#    on a particular system, place or move the name in each
#    of the tables.  

package setup;

#  Set global variables to be retrieved by calling Arch subroutine
#  

#  Table of operating system names in variable $^O.
#  Add another name in one or more categories below as appropriate
#  For example, if your Mac GNUPlot uses X-windows instead of Aqua,
#    replace the `darwin=>"aqua"' line below with `darwin=>"x11"'.
#
#  Some common names:
#    System      $^O string
#      FreeBSD       freebsd 
#      Linux         linux  
#      Mac OS X      darwin
#      openbsd       openbsd 
#      SunOS         solaris
#      Windows       MSWin32
#
#  If your platform is not one of these, you will need to add its
#    name to each one of the tables.  For example if Perl `$O' has
#    the value "weirdOS", and it doesn't support FIFO pipes, then
#    add `weirdOS=>0' to the the %osnames table if it isn't a
#    MS Windows, or `=>1' if it is, and `weirdOS=>0 to the
#    %fifo table.  Or if your Windows
#    does use X11 for GNUPlot, change its line in the %xwin table
#    to `MSWin32=>1' and in the %GNUterm table to `MSWin32=>"x11"'.
#
#  table of OS names in Perl variable $^0 and whether they are MS:
%osnames = (
MSWin32=> 1,
linux=> 0,
freebsd=> 0,
darwin=>0,
openbsd=> 0,
solaris=> 0,
);

#  UNIX-like systems that support FIFO-pipe files: 
%fifo = (
linux=>1,
freebsd=>1,
darwin=>1,
openbsd=>1,
solaris=>1,
#  Others that don't (use this if not sure):=>0,
MSWin32=>0,
);

#  X-windows is the primary windowing system
%xwin = (
linux=>1,
solaris=>1,
#  X-windows not primary (use this if not sure)
freebsd=>0,
darwin=>0,
openbsd=>0,
MSWin32=>0,
);

#  terminal type to be used for GNUPlot 
#    (run GNUPlot interactively and type `set terminal' for a list)
%GNUterm = (
linux=>"x11",
darwin=>"none",  #might be "aqua" or "x11"; let GNUPlot decide
solaris=>"x11",
MSWin32=>"windows",
);

sub isitMSW {
  my ($a);
  $a = $osnames{$^O};
  return $a if (defined($a));
  warn "OS platform '$^O' is not in `setup.pm' module (Default NOT MS Windows)" unless(defined($a));
  return 0;
}

sub fifo {
  my($a);
  $a = $fifo{$^O};
  if (!defined($a)) {
    return 0;
  } else {
    return $a;;
  }
}

sub xwin {
  my($a);
  $a = $xwin{$^O};
  if (!defined($a)) {
    return 0;
  } else {
    return $a;;
  }
}

sub GNUterm {
  my($a);
  $a = $GNUterm{$^O};
  if (!defined($a)) {
    return "none"; #let GNUPlot use its default
  } else {
    return $a;;
  }
}

sub Version {
  #change this line to change version reported by COMP and SYN
  return "v.0.1";
}

return 1;  #module OK value
