package stepfunc;

#used for calculating step-function approximations for components with state

@input_subs = ();
@state_subs = ();

@count = ();

@out_sum = ();
@run_sum = ();
@state_sum = ();

# Initializes step-function approximating
# $_[0] = reference to array with input subdomains
# $_[1] = reference to array with state subdomains
sub init {
  my $subdomain;

  my ($iref, $sref) = @_;
  @input_subs = @$iref;
  @state_subs = @$sref;

  $averages_calculated = 0;
}

# Records a data point and tracks which subdomain the point landed in
#
# first param is "R" for random (state w/sequences)
# $_[1] = input
# $_[2] = initial state
# $_[3] = output
# $_[4] = runtime
# $_[5] = final state
# 
#or "S" for systematic state
#or "C" for concurrent
# next two params are the subdomain indices
#  followed by list above for "S"
#  followed by: input, return from conc, out to conc, final out, three runtimes for "C"
#
sub process {
  my ($dum, $isub, $ssub, $input, $istate, $output, $runtime, $fstate);
  my ($input2, $output1, $output2, $runtime1, $runtime3, $runtime4);
  my ($subdomain, @subd, $c);
  my ($sel) = $_[0];
  if ($sel eq "R") {
    ($dum, $input, $istate, $output, $runtime, $fstate) = @_; 
  } else { #lead with subd indices
    if ($sel eq "S") { #systematic state
      ($dum, $isub, $ssub, $input, $istate, $output, $runtime, $fstate) = @_; 
    } elsif ($sel eq "C") { #concurrent
      ($dum, $isub, $ssub, $input, $input2, $output, $output2, $runtime, $runtime3, $runtime4) = @_; 
#print "$dum, $isub, $ssub, $input, $input2, $output, $output2, $runtime, $runtime3, $runtime4\n"; #debug
    } else {
      die "Bad selector parameter $sel to process()";
    }
  }
#print " $sel: $isub, $ssub, $input, $istate\n";  #debug

  if ($sel eq "R") {
#TBD: These should use binary search!
    # find the subdomain this falls in
    $c = 0;
    INPUT_SUB: foreach $subdomain (@input_subs) {
      @subd = split(' ',$subdomain);
      if ($input >= $subd[0] && $input < $subd[1]) {
        $isub = $c;
        last INPUT_SUB;
      }
      $c++;
    }
    $c = 0;
    STATE_SUB: foreach $subdomain (@state_subs) {
      @subd = split(' ',$subdomain);
      unless (defined($istate)) { #initial undefined state doesn't count
        return 1;
        }
      if ($istate >= $subd[0] && $istate < $subd[1]) {
        $ssub = $c;
        last STATE_SUB;
      }
      $c++;
    }
    return 0 unless ( defined($isub) && defined($ssub) );
  }
  unless (defined ($count[$isub])) {
     $out_sum[$isub][$ssub] = 0;
     $out_sum_sqs[$isub][$ssub] = 0;
     $out2_sum[$isub][$ssub] = 0;
     $out2_sum_sqs[$isub][$ssub] = 0;
     $run_sum[$isub][$ssub] = 0;
     $run_sum_sqs[$isub][$ssub] = 0;
     $run3_sum[$isub][$ssub] = 0;
     $run3_sum_sqs[$isub][$ssub] = 0;
     $run4_sum[$isub][$ssub] = 0;
     $run4_sum_sqs[$isub][$ssub] = 0;
     $state_sum[$isub][$ssub] = 0;
     $state_sum_sqs[$isub][$ssub] = 0;
     $count[$isub][$ssub] = 0;
  }
 
  $out_sum[$isub][$ssub] += $output;
  $out_sum_sqs[$isub][$ssub] += $output*$output;
  $run_sum[$isub][$ssub] += $runtime;
  $run_sum_sqs[$isub][$ssub] += $runtime*$runtime;
  if ($sel eq "C") {
    $out2_sum[$isub][$ssub] += $output2;
    $out2_sum_sqs[$isub][$ssub] += $output2*$output2;
    $run3_sum[$isub][$ssub] += $runtime3;
    $run3_sum_sqs[$isub][$ssub] += $runtime3*$runtime3;
    $run4_sum[$isub][$ssub] += $runtime4;
    $run4_sum_sqs[$isub][$ssub] += $runtime4*$runtime4;
  } else { #state
    $state_sum[$isub][$ssub] += $fstate;
    $state_sum_sqs[$isub][$ssub] += $fstate*$fstate;
  }
  $count[$isub][$ssub]++;
  return 1;
}

sub create_splot{
 ($Cflag, $name) = @_; # $Cflag true for concurrent
                          # $name of the data array, file, etc., e.g., "run3".
 unless ($averages_calculated) { &calculate_averages($Cflag); }

 #get x, y, and z values for plateaus
 open THEFILE, ">" . $name . "_step.dat";
 $i = 0;
 $j1 = 0;
 $j2 = 0;
 BIGLOOP: while (1) {
  nextISUB:
  if ($j2 > $#state_subs) {
    $i++;
    if (!defined($count[$i]) && $i <= $#input_subs) {
      goto nextISUB;
    } elsif ($i > $#input_subs) {
      last BIGLOOP;
    }
    $j1 = 0;
    $j2 = 0;
  }

  @isub = split(' ', $input_subs[$i]);

  $x = $isub[0];
  $j1 = halfplat($i, $j1);
  print THEFILE "\n";

  $x = $isub[1];
  $j2 = halfplat($i, $j2);
  print THEFILE "\n\n";
 }
 close THEFILE;

}

# Creates splot-compatible data files for plotting the plateaus
# $_[0] = file name for output plateaus
# $_[1] = file name for runtime plateaus
# $_[2] = file name for final state plateaus
sub create_splot_files {
 my ($out_file, $run_file, $state_file) = @_; #subroutine arguments

 unless ($averages_calculated) { &calculate_averages(0); }

 #get x, y, and z values for plateaus
 open OUT, ">".$out_file;
 open RUN, ">".$run_file;
 open STATE, ">".$state_file; 
 $i = 0;
 $j1 = 0;
 $j2 = 0;
 BIGLOOP: while (1) {
  nextISUB:
  if ($j2 > $#state_subs) {
    $i++;
    if (!defined($count[$i]) && $i <= $#input_subs) { 
      goto nextISUB;
    } elsif ($i > $#input_subs) {
      last BIGLOOP;
    }
    $j1 = 0;
    $j2 = 0;
  }

  @isub = split(' ', $input_subs[$i]);

  $x = $isub[0];
  $j1 = halfplateau($i, $j1);
  print OUT "\n";
  print RUN "\n";
  print STATE "\n";

  $x = $isub[1];
  $j2 = halfplateau($i, $j2);
  print OUT "\n\n";
  print RUN "\n\n";
  print STATE "\n\n";
 }
 close OUT;
 close RUN;
 close STATE;
}

# Creates .ccft file and corresponding .ccfc file
# $_[0] = file name for ccft file
# $_[1] = first-line description
sub create_ccft_files {
 my ($ccf_file, $name_opts) = @_; #subroutine arguments
 $ccfc_file = $ccf_file."c";
 # ccft can likely go away when it's safe...
 $ccft_file = $ccf_file."t";
 @head = split(' ',$name_opts);

 unless ($averages_calculated) { &calculate_averages($head[0] eq "concurrent"); }

 open CCFT, ">".$ccft_file;
 open CCFC, ">$ccfc_file";

 print CCFT $name_opts."\n";
 print CCFC "$head[0] 1 " . scalar(@input_subs) . " " . scalar(@state_subs) . " " . "$head[3] $head[2]\n"; # "state/concurrent 1" insubs stsubs initst "R/U"
 for my $i (0..$#input_subs) {
   @subdline = split(' ',$input_subs[$i]);
   print CCFC "$subdline[0] $subdline[1]\n";
 }
 for my $i (0..$#state_subs) {
   @subdline = split(' ',$state_subs[$i]);
   print CCFC "$subdline[0] $subdline[1]\n";
 }

 STATE: for my $j (0..$#state_subs) { #loop through state subdomains
  INPUT: for my $i (0..$#input_subs) { #loop through input subdomains
   if ($count[$i][$j]) {
     if ($head[0] eq "state") {
       print CCFT "$count[$i][$j] $out_avg[$i][$j] $run_avg[$i][$j] $state_avg[$i][$j] $out_err[$i][$j] $run_err[$i][$j] $state_err[$i][$j] ";
       print CCFC "$count[$i][$j] $out_avg[$i][$j] $run_avg[$i][$j] $state_avg[$i][$j] $out_err[$i][$j] $run_err[$i][$j] $state_err[$i][$j]\n";
     } elsif ($head[0] eq "concurrent") {
#print "1:$count[$i][$j] 2:$out_avg[$i][$j] 3:$out2_avg[$i][$j] 4:$run_avg[$i][$j] 5:$run3_avg[$i][$j] 6:$run4_avg[$i][$j] 7:$out_err[$i][$j] 8:$out2_err[$i][$j] 9:$run_err[$i][$j] 10:$run3_err[$i][$j] 11:$run4_err[$i][$j]\n"; #debug
       print CCFC "$count[$i][$j] $out_avg[$i][$j] $out2_avg[$i][$j] $run_avg[$i][$j] $run3_avg[$i][$j] $run4_avg[$i][$j] $out_err[$i][$j] $out2_err[$i][$j] $run_err[$i][$j] $run3_err[$i][$j] $run4_err[$i][$j]\n";
     }
   } else {
     print CCFT "0 0 0 0 0 0 0 ";
     if ($head[0] eq "state") {
       print CCFC "0 0 0 0 0 0 0\n"
     } elsif ($head[0] eq "concurrent") {
       print CCFC "0 0 0 0 0 0 0 0 0 0 0\n";
     }
   }
  }
  print CCFT "\n";
 }
 close CCFT;
 close CCFC;
}

# Prints an error table
# $_[0]: string "Output" or "Runtime" or "Final State", which table to print
sub ErrorTable{
my ($which) = $_[0];
my ($i, $j, $s, @subd);
#debug
#print "|$which|\n";
print "Relative % r-m-s errors in $which \n";
print "	";
foreach $s (@input_subs) {
  @subd = split(" ",$s);
  #printf "[%2.1f,%2.1f)	", $subd[0],$subd[1];
  printf "[%2.1f	", $subd[0]; #abbreviate the [ , ) to [ 
}
print "\n";
$totArea = 0;
$totErr = 0;
$maxErr = 0;
for my $j (0..$#state_subs) {
  @ssubd = split(" ",$state_subs[$j]);
  #printf "[%2.1f,%2.1f)	", $ssubd[0],$ssubd[1];
  printf "[%2.1f	", $ssubd[0]; #abbreviate the [ , ) to [ 
  for my $i (0..$#input_subs) {
    $thisArea = ($subd[1]-$subd[0])*($ssubd[1]-$ssubd[0]);
    if ($count[$i][$j]) {
      if ($which eq "Output") {
	$thisErr = $out_err[$i][$j]/sqrt($out_sqs/$rms_count);
      } elsif ($which eq "Runtime") {
	$thisErr = $run_err[$i][$j]/sqrt($run_sqs/$rms_count);
      } elsif ($which eq "Finalstate") { 
	$thisErr = $state_err[$i][$j]/sqrt($state_sqs/$rms_count);
      } elsif ($which eq "Parallelruntime") { 
	$thisErr = $run3_err[$i][$j]/sqrt($run3_sqs/$rms_count);
      } elsif ($which eq "Lastruntime") { 
	$thisErr = $run4_err[$i][$j]/sqrt($run4_sqs/$rms_count);
      } elsif ($which eq "Parallelinput") { 
	$thisErr = $out2_err[$i][$j]/sqrt($out2_sqs/$rms_count);
      } else {
        die "bad type of error table";
      }
      $totArea += $thisArea;
      printf "%3.1f	", 100.0*$thisErr;
      $totErr += $thisErr*$thisArea;
      if ($thisErr > $maxErr) {
        $maxErr = $thisErr;
      }
    }
    else {
      print "**	";
    }
  }
print "\n";
}  
if ($totArea > 0) {
  $totErr *= 100.0/$totArea;
  $maxErr *= 100.0;
  printf "Weighted r-m-s (max) error: %4.2f%% (%4.2f%%)\n",$totErr,$maxErr;
$RMS = sqrt($out_sqs/$rms_count); print "normalized to $RMS total r-m-s on $rms_count values\n"; #debug
}
return 1;
}

############### END PUBLIC SUBROUTINES

# $_0 = count of points
# $_1 = sum of points
# $_2 = sum of squares of points
# Returns almost the relative std. dev.
sub rmsError { 
my ($n, $T, $sq) = @_;
#print "$n, ave $T, sqs $sq\n";  if ($sq == 0) {return 0;} #debug
my ($ave, $eps);
if ($n == 0 || $sq == 0) {return 0;}
$ave = $T/$n;
#$eps = .000000001; #so as not to div by zero:  kludge
#return sqrt(abs($sq - $ave*$ave*$n)/$n)/(abs($ave)+$eps);  #wrong!
#return sqrt(1 - $n*$ave*$ave/$sq);  #wrong because normalized to subd only
return sqrt(abs($sq/$n - $ave*$ave)); #normalize later
}

# calculate_averages - used internally to calculate the averages after all data has been collected
sub calculate_averages {
 my $concurrent = $_[0];  #parameter is switch, otherwise "state"
 $rms_count = 0;
 $out_sqs = 0;
 $run_sqs = 0;
 $out2_sqs = 0;
 $run3_sqs = 0;
 $run4_sqs = 0;
 $state_sqs = 0;
 INPUT: for my $i (0..$#count) { #loop through input subdomains
  unless (defined($count[$i])) { next INPUT; }
  STATE: for my $j (0..@{$count[$i]}) { #loop through state subdomains
   unless ($count[$i][$j]) { next STATE; }
   $rms_count++;
   $out_avg[$i][$j] = $out_sum[$i][$j] / $count[$i][$j];
   $out_sqs += $out_sum_sqs[$i][$j];
   $out_err[$i][$j] = &rmsError($count[$i][$j],$out_sum[$i][$j],$out_sum_sqs[$i][$j]);
   $run_avg[$i][$j] = $run_sum[$i][$j] / $count[$i][$j];
   $run_sqs += $run_sum_sqs[$i][$j];
   $run_err[$i][$j] = &rmsError($count[$i][$j],$run_sum[$i][$j],$run_sum_sqs[$i][$j]);
   if ($concurrent) {
     $out2_avg[$i][$j] = $out2_sum[$i][$j] / $count[$i][$j];
     $out2_sqs += $out2_sum_sqs[$i][$j];
     $out2_err[$i][$j] = &rmsError($count[$i][$j],$out2_sum[$i][$j],$out2_sum_sqs[$i][$j]);
     $run3_avg[$i][$j] = $run3_sum[$i][$j] / $count[$i][$j];
     $run3_sqs += $run3_sum_sqs[$i][$j];
     $run3_err[$i][$j] = &rmsError($count[$i][$j],$run3_sum[$i][$j],$run3_sum_sqs[$i][$j]);
     $run4_avg[$i][$j] = $run4_sum[$i][$j] / $count[$i][$j];
     $run4_sqs += $run4_sum_sqs[$i][$j];
     $run4_err[$i][$j] = &rmsError($count[$i][$j],$run4_sum[$i][$j],$run4_sum_sqs[$i][$j]);
   } else {
     $state_avg[$i][$j] = $state_sum[$i][$j] / $count[$i][$j];
     $state_sqs += $state_sum_sqs[$i][$j];
     $state_err[$i][$j] = &rmsError($count[$i][$j],$state_sum[$i][$j],$state_sum_sqs[$i][$j]);
   } 
  }
 }
 $averages_calculated = 1;
}

sub halfplat {
  my $i = $_[0];
  my $z = $_[1];
  my $j;

  init:
  if (!defined($count[$i][$z]) && $z <= $#state_subs) { $z++; goto init; }

  SSUB2: for ($j = $z; $j <= $z; $j++) {
    unless (defined($count[$i][$j])) { return $j; }
    @ssub = split(' ', $state_subs[$j]);
    $y = $ssub[0];
    print THEFILE "$x $y ". ${$name."_avg"}[$i][$j] . "\n";
    $y = $ssub[1];
    print THEFILE "$x $y ". ${$name."_avg"}[$i][$j] . "\n";
    #print OUT "$x $y $out_avg[$i][$j]\n";
  }
return $j;
}

# halfplateau - used internally by create_splot_files to help order the point data correctly in
# the splot files so that the data is plotted correctly as square plateaus
sub halfplateau {
  my $i = $_[0];
  my $z = $_[1];
  my $j;

  init:
  if (!defined($count[$i][$z]) && $z <= $#state_subs) { $z++; goto init; }

  SSUB2: for ($j = $z; $j <= $z; $j++) {
    unless (defined($count[$i][$j])) { return $j; }
    @ssub = split(' ', $state_subs[$j]);
    $y = $ssub[0];
    print OUT "$x $y $out_avg[$i][$j]\n";
    print RUN "$x $y $run_avg[$i][$j]\n";
    print STATE "$x $y $state_avg[$i][$j]\n";
    $y = $ssub[1];
    print OUT "$x $y $out_avg[$i][$j]\n";
    print RUN "$x $y $run_avg[$i][$j]\n";
    print STATE "$x $y $state_avg[$i][$j]\n";
  }
return $j;
}
return 1;

