#! /usr/bin/perl -w
# trace utility module
package trace;
##########################################################
# init()
#  no parameters
#  reads parameter information for all components into memory
#  initializes counts for all subdomains of all components to 0
sub init {
  open (P, "<system.pscf") or die "system.pscf missing";
  $ccfile = <P>;  #discard Polish line
  $nc = 0;
  while ($ccfile = <P>) { #read all .ccf file lines
    open (CCF, "<$ccfile") or die "$ccfile unaccountably missing";
    @base = split(/\./,$ccfile);
    $ccfbase = $base[0];
    $listofccf[$nc++] = $ccfbase;  #save list of components (base)
    $sub = <CCF>;  #discard header line
    @subf = split(' ',$sub);
    ${$ccfbase."type"} = $subf[1]; #state|concurrent|(empty)
    $type = $subf[1];
    $blankseen = 0;
    $subcount = 0; #count of subdomain being processed
    while ($sub = <CCF>) { #subdomain lines; store LB & UB
      if ($sub eq "\n") {
        $blankseen = 1;
        $subcount = 0; #start over
      } else {
        @subs = split(" ",$sub);
        if ($blankseen) { #doing states
          ${$ccfbase."StLB"}[$subcount] = $subs[0];
          ${$ccfbase."StUB"}[$subcount] = $subs[1];
        } else { #doing inputs
          ${$ccfbase."InLB"}[$subcount] = $subs[0];
          ${$ccfbase."InUB"}[$subcount] = $subs[1];
        }
        $subcount++;
      }
    }
    close CCF;
    for ($i=0; $i<scalar(@{$ccfbase."InLB"}); $i++) {
      if (defined($type)) { #state or concurrent
        if ($type eq "state") { #there was state
          for ($j=0; $j<scalar(@{$ccfbase."StLB"}); $j++) {
            ${$ccfbase."Hits"}[$i][$j] = 0; #initialize hits
          }
        } else { #concurrent
          for ($j=0; $j<scalar(@{$ccfbase."StLB"}); $j++) {
            ${$ccfbase."H2ts"}[$j] = 0;
          }
        }
      } else { #stateless
        ${$ccfbase."Hits"}[$i] = 0;
      }
    }
  } #end all .ccf files
  close P;
#open (DB,">/tmp/bumped");  #debug  see DB below
} #end init()
##########################################################
# bump(basename, input[, state])   
#  Increments the count of inputs falling in basename's subdomains
#  Last arg missing for stateless components
#  For concurrent, last arg present (an index) indicates 2nd input domain
#
#  Also make a trace by component?
#
sub bump {
  my $base = shift;
  my $in = shift;
  my $st = shift;
  $t = ${$base."type"};
  $stateless = 0;
  $concurrent = 0;
  if (defined($t)) {
    $whatnot = "state";
    if ($t eq "concurrent") {
      $concurrent = 1;
      $whatnot = "parallel input";
    }
  } else {
  $stateless = 1;
  }
#print DB "($in,$st)\n" if (defined($st)); #debug see open DB above
  #find subdomain(s)
  $foundin = -1;
  $foundst = -1;
  if ($concurrent && defined($st)) {
    $st = $in;
    goto ST;
  }
  IN:
  for ($i=0; $i<scalar(@{$base."InLB"}); $i++) {
    if (${$base."InLB"}[$i] <= $in && ${$base."InUB"}[$i] > $in) {
      $foundin = $i;
      last IN;
    } 
  }
  if ($foundin == -1) {
    warn "didn't find a subdomain for input $in in component $base" ;
    return;
  }
  if ($stateless || $concurrent) { #no state
    ${$base."Hits"}[$foundin]++;
    return;
  } 
  ST:
  for ($i=0; $i<scalar(@{$base."StLB"}); $i++) {
    if (${$base."StLB"}[$i] <= $st && ${$base."StUB"}[$i] > $st) {
      $foundst = $i;
      last ST;
    }
  }
  if ($foundst == -1) {
    warn "didn't find a subdomain for $whatnot $st in component $base" ;
    return;
  }
  if ($concurrent) {
    ${$base."H2ts"}[$foundst]++;
  } else {
    ${$base."Hits"}[$foundin][$foundst]++;
  }
  return;
} #end bump()
#########################################################
# done()
#  puts total counts in "trace_report" file for all components
#
sub done {
  open(F, ">trace_report");
  for ($i=0; $i<$nc; $i++) { #do each component in list
    $base = $listofccf[$i];
    $type = ${$base."type"};
    print F "Component described by $base.ccf";
    #if (scalar(@{$base."StLB"})) { #has state
    if ($type eq "state") { #has
      print F " w/state\n\n";
      for ($j=0; $j<scalar(@{$base."InLB"}); $j++) {
        print F qq{input [${$base."InLB"}[$j], ${$base."InUB"}[$j]) states:\n};
        for ($js=0; $js<scalar(@{$base."StLB"}); $js++) {
          print F qq{	[${$base."StLB"}[$js], ${$base."StUB"}[$js]) ${$base."Hits"}[$j][$js]\n};
        }
        print F "\n";
      }
    } else {
      print F "\n\n";
      for ($j=0; $j<scalar(@{$base."InLB"}); $j++) {
        print F qq{[${$base."InLB"}[$j], ${$base."InUB"}[$j]) ${$base."Hits"}[$j]\n};
      }
    }
    if ($type eq "concurrent") { #make a 2nd entry
      print F "\n\n";
      print F "Component described by $base.ccf parallel input\n\n";
      for ($j=0; $j<scalar(@{$base."StLB"}); $j++) {
        print F qq{[${$base."StLB"}[$j], ${$base."StUB"}[$j]) ${$base."H2ts"}[$j]\n};
      }
    }
    print F "\n\n";
  }
  close F;
} #end done()
#########################################################
return 1;
